#!/usr/bin/env python
# -*- coding: utf-8 -*-

import logging

ASR_DEBUG = False

# Debug logger
_logger = logging.getLogger(__name__)


class T_I_Kennlinie(object):
    """!
    Bildet die Kennlinie für die Adaptive Stromregelung ab.
    """
    def __init__(self, t1, i1, t2, i2):
        """!
        Die Kennlinie wird mit den beiden Stützstellen initialisiert.
        @param t1: Temperatur Stützstelle 1
        @param i1: Strom Stützstelle 1
        @param t2: Temperatur Stützstelle 2
        @param i2: Strom Stützstelle 2
        """
        _logger.debug("T_I_Kennlinie init")
        self._name = "T_I_Kennlinie"

        # Steigung der linearen Begrenzung berechnen
        self._m = (i2 - i1) / (t2 - t1)
        # y-Achsen Schnittpunkt der linearen Begrenzung berechnen
        self._b = i2 - t2 * self._m
        # Maximaler Strom entspricht Wert von i1
        self._i_max = i1

    def __call__(self, t_ist):
        """!
        Gibt den maximal erlaubten Strom in Abhängigkeit der Temperatur zurück.
        Bis t1 ist dieser Wert i1 (i_max), über t1 greift eine lineare Begrenzung.
        Diese wird anhand der Stützstellen (t1, i1) und (t2, i2) definiert.
        @param t_ist: Temperatur, für die der maximal erlaubte Strom berechnet werden soll
        @return Maximal erlaubter Strom bei t_ist
        """
        # Lineare Begrenzung
        i = self._m * t_ist + self._b
        # Entweder i_max oder i zurückgeben, je nachdem welcher kleiner ist.
        return min(self._i_max, i)


class AdaptiveStromregelung(object):
    """!
    Adaptive Stromregelung
    Begrenzt den (eff.) Strom durch die Heizelemente in dem sie die Leistungsanforderung
    in Abhängigkeit der Brennkammertemperatur begrenzt.
    """
    def __init__(self, t1, i1, t2, i2):
        """!
        Die Stromregelung wird mit den beiden Stützstellen für die Temperatur-Strom-Kennlinie initialisiert.
        @param t1: Temperatur Stützstelle 1
        @param i1: Strom Stützstelle 1
        @param t2: Temperatur Stützstelle 2
        @param i2: Strom Stützstelle 2
        """
        _logger.debug("AdaptiveStromregelung init")
        self._name = "AdaptiveStromregelung"
        self._t_i_kl = T_I_Kennlinie(t1, i1, t2, i2)
        self._lim = False
        self._p_lim = 100
        self._i_ist_mean = 0

    def __call__(self, t_ist, i_ist, p_ist, p_req):
        """!
        Berechnet die maximal erlaubte Leistungsanforderung für die Heizelemente.
        @param t_ist: Aktuelle Ist-Temperatur der Brennkammer/Heizelemente.
        @param i_ist: Aktueller Strom (Effektivwert) durch die Heizelemente.
        @param p_ist: Aktueller Leistungswert der Heizelemente/Leistungseinheit.
        @param p_req: Gewünschte Leistungsanforderung an die Heizelemente/Leistungseinheit.
        @return Begrenzte Leistungsanforderung für die Heizelemente/Leistungseinheit.
        """
        i_max = self._t_i_kl(t_ist)                     # Maximal erlaubter Strom bei Ist-Temperatur laut Kennlinie
        
        # self._i_ist_mean -= self._i_ist_mean / 4        # einfache Mittelwertbildung
        # self._i_ist_mean += i_ist / 4

        if i_ist > i_max:        
            self._p_lim = min(self._p_lim, p_req)       # beim ersten Ansprechen in den Bereich stellen
            self._p_lim -= 0.1                          # kleine Schritte zum Ausmitteln
            self._lim = True
            
        else:
            if self._p_lim < 100:
                self._p_lim += 0.1                      # langsam Aufregeln
                self._lim = False
       
        if ASR_DEBUG:
            _logger.debug("{}C, {}A(ist), {}A(max), {}%(ist), {}%(soll), {}%(lim)".format(t_ist, i_ist, i_max, p_ist, p_req, p_lim))

        return min(p_req, self._p_lim)

    @property
    def lim(self):
        """!
        Funktion zum Prüfen, ob Strombegrenzer aktiv oder nicht
        @return Status des Strombegrenzers.
        """
        return self._lim
