#!/usr/bin/env python
# -*- coding: utf-8 -*-

import logging
import subprocess
import sys
from platform import system

import mraa

_logger = logging.getLogger(__name__)
ENABLE_DEBUG_LOG = False


class RV8523(object):
    dev = None

    I2CDEV = 0x0        # I2C Bus
    I2CADDR = 0x68      # RTC Chip Adresse

    # Registeradressen
    CONTROL1 = 0x00
    CONTROL2 = 0x01
    CONTROL3 = 0x02
    SECONDS = 0x03
    MINUTES = 0x04
    HOURS = 0x05
    DAYS = 0x06
    WEEKDAYS = 0x07
    MONTHS = 0x08
    YEARS = 0x09
    MINUTEALARM = 0x0A
    HOURALARM = 0x0B
    DAYALARM = 0x0C
    WEEKDAYALARM = 0x0D
    FREQUENCYOFFSET = 0x0E
    TIMERCLKOUT = 0x0F
    TIMERACLCK = 0x10
    TIMERA = 0x11
    TIMERBCLK = 0x12
    TIMERB = 0x13

    def __init__(self):
        RV8523.dev = mraa.I2c(RV8523.I2CDEV)
        RV8523.dev.frequency(mraa.I2C_STD)
        RV8523.dev.address(RV8523.I2CADDR)
        RV8523.dev.writeWordReg(0x00, 0xff00)

    def writeReg(self, regaddr, regvalue):
        """!
        Beschreibt das RTC Register mit Adresse regaddr mit dem Wert regvalue.
        @param regaddr: RTC Register Adresse
        @param regvalue: Wert, der in das RTC Register geschrieben werden soll
        """
        RV8523.dev.address(RV8523.I2CADDR)
        RV8523.dev.writeReg(regaddr, regvalue)

    def readReg(self, regaddr):
        """!
        Ließt das RTC Register mit Adresse regaddr und gibt dessen Wert zurück.
        @param regaddr: RTC Register Adresse
        @return Wert des gelesenen Registers
        """
        RV8523.dev.address(RV8523.I2CADDR)
        return RV8523.dev.readReg(regaddr)


class RTC(object):
    """!
    Dieses Modul ist für die Ansteuerung des RTC-Chips auf der Hardware zuständig.
    """

    def __init__(self):
        if ENABLE_DEBUG_LOG:
            _logger.debug("RTC init")
        self._name = "RTC"
        self._rtcDev = RV8523()

    def setDate(self, year, month, day, update_systemdate=False):
        """!
        Mit dieser Funktion kann das Datum des RTC-Chips geändert werden. Die Parameter year, month und day
        entsprechen dem Jahr, Monat und Tag und werden als Integer übergeben. Diese werden in der Funktion
        in das korrekte RTC Format umgewandelt und in die jeweiligen Hardware-Register übertragen.
        @param year: Jahr (0..99)
        @param month: Monat (1..12)
        @param day: Tag (1..31)
        @param update_systemdate: Wenn True, wird die Linux-Systemdatum gesetzt
        """

        # year, month und day in BCD konvertieren und auf ein Byte begrenzen.
        # https://forum.inductiveautomation.com/t/convert-int-to-bcd-data/11697
        bcd_dd = int(str(day), 16) & 0xFF
        bcd_mm = int(str(month), 16) & 0xFF
        bcd_yy = int(str(year), 16) & 0xFF
        if ENABLE_DEBUG_LOG:
            _logger.debug("Date BCD conversion done (YY, MM, DD): {} {} {}".format(hex(bcd_yy), hex(bcd_mm), hex(bcd_dd)))

        # BCD konvertiertes Datum in RTC schreiben
        ctrl1 = self._rtcDev.readReg(RV8523.CONTROL1)       # Control Register 1 lesen
        ctrl1 = ctrl1 | 0x20                                # STOP Bit setzen
        self._rtcDev.writeReg(RV8523.CONTROL1, ctrl1)       # Control Register 1 schreiben, stoppt RTC
        self._rtcDev.writeReg(RV8523.DAYS, bcd_dd)
        self._rtcDev.writeReg(RV8523.MONTHS, bcd_mm)
        self._rtcDev.writeReg(RV8523.YEARS, bcd_yy)
        ctrl1 = self._rtcDev.readReg(RV8523.CONTROL1)       # Control Register 1 lesen
        ctrl1 = ctrl1 & 0xDF                                # STOP Bit löschen
        self._rtcDev.writeReg(RV8523.CONTROL1, ctrl1)       # Control Register 1 schreiben, startet RTC wieder

        # Linux-Systemzeit setzen
        if update_systemdate:
            rtc_date = "{:02d}-{:02d}-{:02d}".format(year, month, day)
            subprocess.call(['date', '+%y%m%d', '-s', rtc_date])

    def getDate(self):
        """!
        Diese Funktion liest das Datum aus dem RTC-Chip und gibt dieses als String zurück.
        Der String hat das Format YY-MM-DD, also Jahr (nur 2 Stellen), Monat und Tag.
        @return RTC Datum als String im Format YY-MM-DD
        """
        if system() == 'Windows':
            value = subprocess.check_output("date /t", shell=True)
            int_yy = int(value[0:2])
            int_mm = int(value[3:5])
            int_dd = int(value[8:10])
        else:
            # Datum aus RTC lesen
            bcd_dd = self._rtcDev.readReg(RV8523.DAYS) & 0x3F
            bcd_mm = self._rtcDev.readReg(RV8523.MONTHS) & 0x1F
            bcd_yy = self._rtcDev.readReg(RV8523.YEARS) & 0xFF
            if ENABLE_DEBUG_LOG:
                _logger.debug("Read RTC date (YY, MM, DD): {} {} {}".format(hex(bcd_yy), hex(bcd_mm), hex(bcd_dd)))

            # BCD nach Int konvertieren
            # https://forum.inductiveautomation.com/t/convert-int-to-bcd-data/11697
            try:
                int_yy = int('%x' % bcd_yy)
                int_mm = int('%x' % bcd_mm)
                int_dd = int('%x' % bcd_dd)
            except ValueError:
                _logger.error('can not convert RTC date {} {} {}'.format(bcd_dd, bcd_mm, bcd_yy))
                int_yy = 0
                int_mm = 1
                int_dd = 1
                self.resetRtc()

        # Konvertiertes Datum zurückgeben
        return "{:02d}-{:02d}-{:02d}".format(int_yy, int_mm, int_dd)

    def setTime(self, hour, minute, second, update_systemtime=False):
        """!
        Mit dieser Funktion kann die Uhrzeit des RTC-Chips geändert werden. Die Parameter hour, minute und second
        entsprechen der Stunde, Minute und Sekunde und werden als Integer übergeben. Diese werden in der Funktion
        in das korrekte RTC Format umgewandelt und in die jeweiligen Hardware-Register übertragen. Die Stunde wird
        im 24h Format übergeben (also 00..23).
        @param hour: Stunde im 24h Format (0..23)
        @param minute: Minute (0..59)
        @param second: Sekunde (0..59)
        @param update_systemtime: Wenn True, wird die Linux-Systemzeit gesetzt
        """

        # hour, minute und second in BCD konvertieren und auf ein Byte begrenzen.
        # https://forum.inductiveautomation.com/t/convert-int-to-bcd-data/11697
        bcd_ss = int(str(second), 16) & 0xFF
        bcd_mm = int(str(minute), 16) & 0xFF
        bcd_hh = int(str(hour), 16) & 0xFF
        if ENABLE_DEBUG_LOG:
            _logger.debug("Time BCD conversion done (HH, MM, SS): {} {} {}".format(hex(bcd_hh), hex(bcd_mm), hex(bcd_ss)))

        # BCD konvertierte Zeit in RTC schreiben
        ctrl1 = self._rtcDev.readReg(RV8523.CONTROL1)       # Control Register 1 lesen
        ctrl1 = ctrl1 | 0x20                                # STOP Bit setzen
        self._rtcDev.writeReg(RV8523.CONTROL1, ctrl1)       # Control Register 1 schreiben, stoppt RTC
        self._rtcDev.writeReg(RV8523.SECONDS, bcd_ss)
        self._rtcDev.writeReg(RV8523.MINUTES, bcd_mm)
        self._rtcDev.writeReg(RV8523.HOURS, bcd_hh)
        ctrl1 = self._rtcDev.readReg(RV8523.CONTROL1)       # Control Register 1 lesen
        ctrl1 = ctrl1 & 0xDF                                # STOP Bit löschen
        self._rtcDev.writeReg(RV8523.CONTROL1, ctrl1)       # Control Register 1 schreiben, startet RTC wieder

        # Linux-Systemzeit setzen
        if update_systemtime:
            rtc_time = "{:02d}:{:02d}:{:02d}".format(hour, minute, second)
            subprocess.call(['date', '+%T', '-s', rtc_time])

    def getTime(self):
        """!
        Diese Funktion liest die Uhrzeit aus dem RTC-Chip und gibt dieses als String zurück. Der String hat das
        Format HH:MM:SS, also Stunde, Minute und Sekunde. Die Stunde wird im 24h Format ausgegeben (also 00..23).
        @return RTC Uhrzeit als String im Format HH:MM:SS
        """
        if system() == 'Windows':
            value = subprocess.check_output("time /t", shell=True)
            int_hh = int(value[0:2])
            int_mm = int(value[3:5])
            int_ss = 0
        else:
            # Datum aus RTC lesen
            bcd_ss = self._rtcDev.readReg(RV8523.SECONDS) & 0x7F     # 0x7F filtert das 'RTC invalid' Flag
            bcd_mm = self._rtcDev.readReg(RV8523.MINUTES) & 0x7F
            bcd_hh = self._rtcDev.readReg(RV8523.HOURS) & 0x3F
            if ENABLE_DEBUG_LOG:
                _logger.debug("Read RTC time (HH, MM, SS): {} {} {}".format(hex(bcd_hh), hex(bcd_mm), hex(bcd_ss)))

            # BCD nach Int konvertieren
            # https://forum.inductiveautomation.com/t/convert-int-to-bcd-data/11697
            try:
                int_hh = int('%x' % bcd_hh)
                int_mm = int('%x' % bcd_mm)
                int_ss = int('%x' % bcd_ss)
            except ValueError:
                _logger.error('can not convert RTC time {} {} {}'.format(bcd_hh, bcd_mm, bcd_ss))
                int_hh = 0
                int_mm = 0
                int_ss = 0
                self.resetRtc()

        # Konvertierte Zeit zurückgeben
        return "{:02d}:{:02d}:{:02d}".format(int_hh, int_mm, int_ss)

    def resetRtc(self):
        """!
        Setzt die RTC auf Default-Werte.
        """
        _logger.warning("RTC reset")
        self.setDate(00,01,01)
        self.setTime(01,00,00)
        #self._rtcDev.writeReg(RV8523.CONTROL1, 0x58)  # Control Register 1 mit 0x58 beschreiben.

    def rtcValid(self):
        """!
        Prüft, ob die RTC Zeit gültig ist.
        @return True, wenn die RTC Zeit gültig ist.\n False, wenn die RTC Zeit ungültig ist.
        """
        if system() == 'Windows':
            return True
        else:
            # Wenn Bit 7 gesetzt ist, ist die RTC Zeit ungültig.
            return False if self._rtcDev.readReg(RV8523.SECONDS) & 0x80 == 0x80 else True


if __name__ == '__main__':
    """!
    Kommandozeilen Funktionen z.B. zum Synchronisieren der Systemzeit mit der RTC.
    Das Argument -updateSysTime synchronisiert die Systemzeit mit der RTC.
    """
    from logging.config import dictConfig

    logging_config = dict(
        version=1,
        disable_existing_loggers=False,
        formatters={
            'f': {'format': '%(asctime)s %(name)-12s %(levelname)-8s [%(funcName)s():%(lineno)s]\t %(message)s'}
        },
        handlers={
            'h': {'class': 'logging.StreamHandler',
                  'formatter': 'f',
                  'level': logging.DEBUG}
        },
        root={
            'handlers': ['h'],
            'level': logging.DEBUG,
        },
    )
    dictConfig(logging_config)

    rtc = RTC()

    rtcDate = rtc.getDate()
    rtcTime = rtc.getTime()
    rtcValid = rtc.rtcValid()

    if (len(sys.argv) == 2) and (sys.argv[1] == '-updateSysTime'):  # Parameter -updateSystemTime synchronisiert die Systemzeit mit der RTC
        if rtcValid:
            subprocess.call(['date', '+%y%m%d', '-s', rtcDate])
            subprocess.call(['date', '+%T', '-s', rtcTime])
        else:
            print 'Cannot update system time, RTC is invalid: ' + rtcDate + ' ' + rtcTime
    else:                                                           # ohne Parameter wird die aktuelle RTC Zeit ausgegeben
        print rtcDate + ' ' + rtcTime + ', rtcValid=' + str(rtcValid)
